import os.path
import unicodedata

from ipatok.ipa import is_letter, is_tie_bar, replace_substitutes
from ipatok.tokens import normalise
from ipatok import tokenise as tokenise_ipa



"""
Paths to the asjp/data dir and the chart file located there.
"""
DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
CHART_FILE = os.path.join(DATA_DIR, 'chart')



class Chart:
	"""
	Object that loads and stores the data from the chart file.
	"""

	def __init__(self):
		"""
		Init the instance's properties: a dict mapping IPA symbols to their
		ASJP counterparts and three dicts for the reverse mapping.
		"""
		self.ipa = {}

		self.asjp_letters = {}
		self.asjp_diacritics = {}
		self.asjp_juxta_letters = {}

	def load(self, path):
		"""
		Populate the instance's properties. The data is expected to be read
		from a file which is parsed as follows:

		- empty lines and lines starting with # are ignored;
		- content lines should consist of tab-separated sub-strings, the first
		  being an IPA symbol, the second its ASJP counterpart, the (optional)
		  third a marker;
		- the marker ✓ indicates that the IPA symbol is also the ASJP's
		  counterpart for the purposes of asjp2ipa conversion;
		- the marker + indicates that the second symbol is an ASJP diacritic;
		- the marker = indicates that the correspondence only applies in the
		  context of an ASJP juxtaposition;
		"""
		with open(path, encoding='utf-8') as f:
			for line in map(lambda x: x.strip(), f):
				if not line or line.startswith('#'):
					continue

				line = line.split('\t')

				if len(line) == 2:
					ipa_symbol, asjp_symbol = line
					flag = ''
				elif len(line) == 3:
					ipa_symbol, asjp_symbol, flag = line
				else:
					continue

				self.ipa[ipa_symbol.replace('ə͡'[1], '')] = asjp_symbol

				if flag == '✓':
					self.asjp_letters[asjp_symbol] = ipa_symbol
				elif flag == '+':
					self.asjp_diacritics[asjp_symbol] = ipa_symbol
				elif flag == '=':
					self.asjp_juxta_letters[asjp_symbol] = ipa_symbol


def convert_ipa_token(token):
	"""
	Convert an IPA token into an ASJP token or raise (Assertion|Index)Error if
	the input does not constitute a valid IPA token.

	Helper for ipa2asjp(ipa_seq).
	"""
	output = []
	has_tie_bar = False

	for char in token:
		if is_letter(char):
			if has_tie_bar:
				affricate = output[-1] + char
				if affricate in chart.ipa:
					output[-1] = chart.ipa[affricate]
				has_tie_bar = False
			else:
				for asjp_char in chart.ipa[char]:
					output.append(asjp_char)

		elif is_tie_bar(char):
			has_tie_bar = True

		elif char == 'n̪'[1] and output[-1] == 'n':
			output[-1] = chart.ipa['n̪']

		elif char in chart.ipa:
			asjp_char = chart.ipa[char]
			if asjp_char in chart.asjp_diacritics:
				output[-1] += asjp_char
			else:
				output.append(asjp_char)

	assert 1 <= len(output) <= 3

	if len(output) != 1:
		output.append('~' if len(output) == 2 else '$')

	return ''.join(output)


def ipa2asjp(ipa_seq):
	"""
	Convert an IPA sequence (string or list) into an ASJP sequence of the same
	type. Raise ValueError if the input is not a valid IPA sequence and raise
	TypeError if it is not a sequence. Usage:

	>>> ipa2asjp('zɛmʲa')
	'zEmy~a'
	>>> ipa2asjp(['z', 'ɛ', 'mʲ', 'a'])
	['z', 'E', 'my~', 'a']

	Part of the package's public API.
	"""
	if isinstance(ipa_seq, str):
		output = []

		for ipa_word in ipa_seq.strip().split():
			try:
				output.append(''.join([
					convert_ipa_token(token)
					for token in tokenise_ipa(ipa_word, replace=True)]))
			except (AssertionError, IndexError):
				raise ValueError('invalid IPA word: {}'.format(ipa_word)) from None

		return ' '.join(output)

	else:
		output = []

		for token in ipa_seq:
			token = replace_substitutes(normalise(token))

			if token == '':
				output.append('')
				continue

			try:
				output.append(convert_ipa_token(token))
			except (AssertionError, IndexError) as err:
				raise ValueError('invalid IPA token: {}'.format(token)) from None

		return output


def convert_asjp_token(token):
	"""
	Convert an ASJP token into an IPA token or raise (Assertion|Key)Error if
	the input does not constitute a valid ASJP token.

	Helper for asjp2ipa(asjp_seq).
	"""
	juxtaposed = False
	output = ''

	if token[-1] in '~$':
		juxtaposed = True
		token = token[:-1]
		token = token.replace('Sx', 'ɧ')

	for char in token:
		if juxtaposed and char in chart.asjp_juxta_letters:
			output += chart.asjp_juxta_letters[char]
		elif char in chart.asjp_diacritics:
			output += chart.asjp_diacritics[char]
		elif char == 'ɧ':
			output += char
		else:
			output += chart.asjp_letters[char]

	return output


def asjp2ipa(asjp_seq):
	"""
	Convert an ASJP sequence (string or list) into an IPA sequence of the same
	type. Raise ValueError if the input is not a valid ASJP sequence and raise
	TypeError if it is not a sequence. Usage:

	>>> asjp2ipa('zEmy~a')
	'zɛmʲa'
	>>> asjp2ipa(tokenise('zEmy~a'))
	['z', 'ɛ', 'mʲ', 'a']

	Part of the package's public API.
	"""
	if isinstance(asjp_seq, str):
		output = []

		for asjp_word in asjp_seq.strip().split():
			try:
				output.append(''.join([
					convert_asjp_token(token) for token in tokenise(asjp_word)]))
			except (AssertionError, KeyError):
				raise ValueError('invalid ASJP word: {}'.format(asjp_word)) from None

		return ' '.join(output)

	else:
		output = []

		for token in asjp_seq:
			if token == '':
				output.append('')
				continue

			try:
				output.append(convert_asjp_token(token))
			except (AssertionError, KeyError):
				raise ValueError('invalid ASJP token: {}'.format(token)) from None

		return output


def tokenise_word(string):
	"""
	Tokenise an ASJP string into a list of tokens or raise ValueError if it
	cannot be unambiguously tokenised. The string is assumed to comprise a
	single word, i.e. it should not contain whitespace chars.

	Helper for tokenise(string).
	"""
	tokens = []

	for ix, char in enumerate(string):
		if char in '~$':
			try:
				for _ in range(1 if char == '~' else 2):
					last_token = tokens.pop()
					tokens[-1] += last_token
				tokens[-1] += char
			except IndexError:
				raise ValueError('ambiguous usage of {} ({})'.format(char, unicodedata.name(char)))

		elif char in '"*':
			try:
				tokens[-1] += char
			except IndexError:
				raise ValueError('word-initial {} ({})'.format(char, unicodedata.name(char)))

		elif char in chart.asjp_letters:
			tokens.append(char)

		else:
			raise ValueError('unknown symbol {} ({})'.format(char, unicodedata.name(char)))

	return tokens


def tokenise(string):
	"""
	Tokenise an ASJP string into a list of tokens. Raise ValueError if it
	cannot be unambiguously tokenised and raise TypeError if it is not a
	string. The input may consist of several words, i.e. whitespace-separated
	sub-strings. Usage:

	>>> tokenise('nova zEmy~a')
	['n', 'o', 'v', 'a', 'z', 'E', 'my~', 'a']

	Part of the package's public API.
	"""
	if not isinstance(string, str):
		raise TypeError('string expected')

	words = string.split()
	output = []

	for word in words:
		try:
			output.extend(tokenise_word(word))
		except ValueError as err:
			raise ValueError('cannot tokenise {}: {!s}'.format(word, err)) from None

	return output


"""
Provide for the alternative spelling.
"""
tokenize = tokenise


"""
Load the chart
"""
chart = Chart()
chart.load(CHART_FILE)



from unittest import TestCase

from hypothesis.strategies import composite, lists, sampled_from
from hypothesis import assume, example, given

import ipatok

from asjp.asjp import chart
from asjp import ipa2asjp, asjp2ipa, tokenise



ASJP_JUXTA_LETTERS = ''.join(sorted(chart.asjp_juxta_letters.keys()))
ASJP_LETTERS = ''.join(sorted(chart.asjp_letters.keys()))



@composite
def asjp_tokens(draw):
	"""
	Composite strategy that generates a non-empty ASJP token.
	"""
	suffix = draw(sampled_from('_"*~$'))

	if suffix in '_"*':
		token = draw(sampled_from(ASJP_LETTERS))
		if suffix != '_':
			token += suffix

	else:
		token = draw(sampled_from(ASJP_LETTERS))
		token += draw(sampled_from(ASJP_LETTERS))

		if suffix == '$':
			token += draw(sampled_from(ASJP_LETTERS))

		assume(any([char not in ASJP_JUXTA_LETTERS for char in token]))

		token += suffix

	return token



class ApiTestCase(TestCase):
	"""
	The IPA strings are sourced from NorthEuraLex (bul, che, deu, hin, lez,
	swe), their ASJP counterparts are derived manually.
	"""

	def test_ipa2asjp_strings(self):
		"""
		IPA-compliant strings should be correctly converted to ASJP.
		"""
		self.assertEqual(ipa2asjp(''), '')

		self.assertEqual(ipa2asjp('sɫɤnt͡sɛ'), 'sloncE')
		self.assertEqual(ipa2asjp('zvɛzda'), 'zvEzda')
		self.assertEqual(ipa2asjp('zɛmʲa'), 'zEmy~a')
		self.assertEqual(ipa2asjp('ɔɡɤn'), 'ogon')
		self.assertEqual(ipa2asjp('javʲa sɛ'), 'yavy~a sE')

		self.assertEqual(ipa2asjp('motː'), 'mot')
		self.assertEqual(ipa2asjp('t͡ʃʼenɪɡ'), 'C"enig')
		self.assertEqual(ipa2asjp('bu͡ʊqʼ'), 'buq"')
		self.assertEqual(ipa2asjp('tʼu͡ʊlɡ'), 't"ulg')
		self.assertEqual(ipa2asjp('bu͡ʊt͡s'), 'buc')

		self.assertEqual(ipa2asjp('zɔnə'), 'zon3')
		self.assertEqual(ipa2asjp('vasɐ'), 'vasa')
		self.assertEqual(ipa2asjp('ʃtaɪ̯n'), 'Stain')
		self.assertEqual(ipa2asjp('ɛɐ̯də'), 'Ead3')
		self.assertEqual(ipa2asjp('fɔʏ̯ɐ'), 'foia')

		self.assertEqual(ipa2asjp('ãːkʰ'), 'a*kh~')
		self.assertEqual(ipa2asjp('ɖʰʊ̄̃ɽʰənaː'), 'dh~u*rh~3na')

		self.assertEqual(ipa2asjp('qʼʷetʼ'), 'q"w~et"')
		self.assertEqual(ipa2asjp('wat͡sʼ'), 'wac"')
		self.assertEqual(ipa2asjp('t͡sʼʷelin ttar'), 'c"w~elin ttar')
		self.assertEqual(ipa2asjp('kʼʷenkʼʷ'), 'k"w~enk"w~')

		self.assertEqual(ipa2asjp('ɧɪnː'), 'Sx~in')
		self.assertEqual(ipa2asjp('ɧæːɳa'), 'Sx~Ena')
		self.assertEqual(ipa2asjp('ɔtːɧɪlɪɡ'), 'otSx~ilig')

	def test_ipa2asjp_lists(self):
		"""
		IPA-compliant tokens should be correctly converted to ASJP tokens.
		"""
		self.assertEqual(ipa2asjp([]), [])
		self.assertEqual(ipa2asjp(['']), [''])

		self.assertEqual(ipa2asjp(['s', 'ɫ', 'ɤ', 'n', 't͡s', 'ɛ']), ['s', 'l', 'o', 'n', 'c', 'E'])
		self.assertEqual(ipa2asjp(['z', 'ɛ', 'mʲ', 'a']), ['z', 'E', 'my~', 'a'])

		self.assertEqual(ipa2asjp(['t͡ʃʼ', 'e', 'n', 'ɪ', 'ɡ']), ['C"', 'e', 'n', 'i', 'g'])
		self.assertEqual(ipa2asjp(['b', 'u͡ʊ', 'qʼ']), ['b', 'u', 'q"'])

		self.assertEqual(ipa2asjp(['ʃ', 't', 'a', 'ɪ̯', 'n']), ['S', 't', 'a', 'i', 'n'])
		self.assertEqual(ipa2asjp(['ʃ', 't', 'a͡ɪ̯', 'n']), ['S', 't', 'a', 'n'])

		self.assertEqual(ipa2asjp(['ãː', 'kʰ']), ['a*', 'kh~'])
		self.assertEqual(ipa2asjp(['ɖʰ', 'ʊ̄̃', 'ɽʰ', 'ə', 'n', 'aː']), ['dh~', 'u*', 'rh~', '3', 'n', 'a'])

		self.assertEqual(ipa2asjp(['qʼʷ', 'e', 'tʼ']), ['q"w~', 'e', 't"'])
		self.assertEqual(ipa2asjp(['w', 'a', 't͡sʼ']), ['w', 'a', 'c"'])

		self.assertEqual(ipa2asjp(['ɧ', 'ɪ', 'nː']), ['Sx~', 'i', 'n'])
		self.assertEqual(ipa2asjp(['ɔ', 'tː', 'ɧ', 'ɪ', 'l', 'ɪ', 'ɡ']), ['o', 't', 'Sx~', 'i', 'l', 'i', 'g'])

	def test_ipa2asjp_errors(self):
		"""
		Non-IPA sequences should raise ValueError and non-sequences should
		raise TypeError.
		"""
		for item in [None, True, 42]:
			with self.assertRaises(TypeError):
				ipa2asjp(item)

		for item in ['zEmy~a', 'павел']:
			with self.assertRaises(ValueError):
				ipa2asjp(item)

		for item in [['z', 'E', 'my~', 'a'], ['п']]:
			with self.assertRaises(ValueError):
				ipa2asjp(item)

	def test_asjp2ipa_strings(self):
		"""
		ASJP strings should be converted into IPA-compliant strings.
		"""
		self.assertEqual(asjp2ipa(''), '')

		self.assertEqual(asjp2ipa('sloncE'), 'slont͡sɛ')
		self.assertEqual(asjp2ipa('zvEzda'), 'zvɛzda')
		self.assertEqual(asjp2ipa('zEmy~a'), 'zɛmʲa')
		self.assertEqual(asjp2ipa('yavy~a sE'), 'javʲa sɛ')

		self.assertEqual(asjp2ipa('a*kh~'), 'ãkʰ')
		self.assertEqual(asjp2ipa('dh~u*rh~3na'), 'dʰũrʰəna')

		self.assertEqual(asjp2ipa('q"w~et"'), 'qʼʷetʼ')
		self.assertEqual(asjp2ipa('wac"'), 'wat͡sʼ')
		self.assertEqual(asjp2ipa('c"w~elin ttar'), 't͡sʼʷelin ttar')
		self.assertEqual(asjp2ipa('k"w~enk"w~'), 'kʼʷenkʼʷ')

		self.assertEqual(asjp2ipa('Sx~in'), 'ɧin')
		self.assertEqual(asjp2ipa('otSx~ilig'), 'otɧiliɡ')

	def test_asjp2ipa_lists(self):
		"""
		ASJP tokens should be converted into IPA-compliant tokens.
		"""
		self.assertEqual(asjp2ipa([]), [])
		self.assertEqual(asjp2ipa(['']), [''])

		self.assertEqual(asjp2ipa(['s', 'l', 'o', 'n', 'c', 'E']), ['s', 'l', 'o', 'n', 't͡s', 'ɛ'])
		self.assertEqual(asjp2ipa(['z', 'E', 'my~', 'a']), ['z', 'ɛ', 'mʲ', 'a'])

		self.assertEqual(asjp2ipa(['q"w~', 'e', 't"']), ['qʼʷ', 'e', 'tʼ'])
		self.assertEqual(asjp2ipa(['w', 'a', 'c"']), ['w', 'a', 't͡sʼ'])

	def test_asjp2ipa_errors(self):
		"""
		Non-ASJP sequences should raise ValueError and non-sequences should
		raise TypeError.
		"""
		for item in [None, True, 42]:
			with self.assertRaises(TypeError):
				asjp2ipa(item)

		for item in ['zɛmʲa', 'павел']:
			with self.assertRaises(ValueError):
				asjp2ipa(item)

		for item in [['z', 'ɛ', 'mʲ', 'a'], ['п']]:
			with self.assertRaises(ValueError):
				asjp2ipa(item)

	@given(lists(asjp_tokens()))
	@example(['Cx~'])
	def test_asjp_ipa_asjp(self, tokens):
		"""
		Tokens generated by asjp_tokens() should be correctly recovered after
		being converted to IPA.
		"""
		self.assertEqual(ipa2asjp(asjp2ipa(tokens)), tokens)

	def test_tokenise(self):
		"""
		ASJP strings should be correctly tokenised.
		"""
		self.assertEqual(tokenise(''), [])

		self.assertEqual(tokenise('zEmy~a'), ['z', 'E', 'my~', 'a'])
		self.assertEqual(tokenise('yavy~a sE'), ['y', 'a', 'vy~', 'a', 's', 'E'])

		self.assertEqual(tokenise('q"w~et"'), ['q"w~', 'e', 't"'])
		self.assertEqual(tokenise('c"w~elin ttar'), ['c"w~', 'e', 'l', 'i', 'n', 't', 't', 'a', 'r'])

		self.assertEqual(tokenise('Sx~in'), ['Sx~', 'i', 'n'])
		self.assertEqual(tokenise('otSx~ilig'), ['o', 't', 'Sx~', 'i', 'l', 'i', 'g'])

	def test_tokenise_errors(self):
		"""
		Non-ASJP strings should raise ValueError and non-strings should raise
		TypeError.
		"""
		for item in [None, True, 42]:
			with self.assertRaises(TypeError):
				tokenise(item)

		for item in ['*a', '"p', 'a~', 'aa$', 'a~$']:
			with self.assertRaises(ValueError):
				tokenise(item)

	@given(lists(asjp_tokens()))
	def test_tokenise_generated(self, tokens):
		"""
		Concatenated tokens generated by asjp_tokens() should be correctly
		tokenised back.
		"""
		self.assertEqual(tokenise(''.join(tokens)), tokens)

	def test_tokenise_ipa(self):
		"""
		Tokenising the IPA input or the ASJP output should be equivalent.
		"""
		for string in ['sɫɤnt͡sɛ', 'tʼu͡ʊlɡ', 'fɔʏ̯ɐ', 'ɖʰʊ̄̃ɽʰənaː', 't͡sʼʷelin ttar', 'ɧɪnː']:
			tokens = ipatok.tokenise(string, replace=True)
			self.assertEqual(tokenise(ipa2asjp(string)), ipa2asjp(tokens))



import os.path

from setuptools import setup, find_packages



BASE_DIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(BASE_DIR, 'README.rst')) as f:
	README = f.read()

with open(os.path.join(BASE_DIR, 'asjp/__version__.py')) as f:
	exec(f.read())



setup(
	name = 'asjp',
	version = VERSION,

	description = 'ASJP conversion and tokenisation utils',
	long_description = README,

	url = 'https://github.com/pavelsof/asjp',

	author = 'Pavel Sofroniev',
	author_email = 'pavelsof@gmail.com',

	license = 'MIT',

	classifiers = [
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3',
		'Topic :: Text Processing :: Linguistic'
	],
	keywords = 'ASJP ASJPcode IPA',

	packages = find_packages(),
	package_data = {'asjp': ['data/*']},

	install_requires = ['ipatok >= 0.2'],
	python_requires = '>=3',

	test_suite = 'asjp.tests',
	tests_require = ['hypothesis >= 3.52']
)
